// $Id: NucleotideDictionary.h 967 2017-07-13 12:04:25Z ge $
// Copyright 2006 Gerald Weber
/// \file NucleotideDictionary.h
/// \brief Class which handles conversions of symbols to internal representations
///
/// $Revision: 967 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_NUCLEOTIDEDICTIONARY_H
#define GBC_NUCLEOTIDEDICTIONARY_H "$Id: NucleotideDictionary.h 967 2017-07-13 12:04:25Z ge $"
#include <fstream>
#include <map>
#include <list>
#include <complex>

namespace std {
template<class _Tp>
/// \brief Dummy operator to enable insertion of objects of type complex
/// \attention This is for compilation purposes only
bool operator<(const std::complex<_Tp>& cx1, const std::complex<_Tp>& cx2)
  {return true;}}

namespace gbc {


/// \brief Conversion dictionary between symbols and internal representation of Nucleotide.
///
/// Implements a conversion dictionary for the symbols which are being read
/// from the data files to some internal representation _InternalTp.
/// In a simplest case _InternalTp could be char or int, but it can also be some
/// more elaborated structure.
///
/// The dictionary file is simply a two-column space separated list of conversions.
/// For example for a NucleotideDictionary<int> this could be
/// \verbatim
/// A 1
/// C 2
/// G 3
/// T 4
/// a 1
/// c 2
/// g 3
/// t 4\endverbatim
/// or for NucleotideDictionary<complex<int> >
/// \verbatim
/// A (1,1)
/// C (-1,-1)
/// G (-1,1)
/// T (1,-1)
/// a (1,1)
/// c (-1,-1)
/// g (-1,1)
/// t (1,-1)\endverbatim
/// \attention Before using this class, make sure there exists an extractor for _InternalTp,
/// i.e, that std::ifstream >> _InternalTp exists.
template<class _InternalTp=char, class _IndexTp=char>
class NucleotideDictionary: public std::map<_IndexTp,_InternalTp>
  {
  public:
    typedef _InternalTp internal_type;
    typedef _IndexTp    index_type;
    typedef std::map<_IndexTp,_InternalTp> map_type;
    typedef std::map<_InternalTp,_IndexTp> inverse_map_type;
    typedef std::map<_InternalTp,_InternalTp> complementary_map_type;
    typedef NucleotideDictionary<_InternalTp,_IndexTp> DictionaryType;
    typedef std::list<_IndexTp> char_list_type;

    inverse_map_type inverse_map; ///< Map for efficient inverse lookup
    complementary_map_type complementary_map; ///< Map for symbols that are complementary

    /// \brief Void constructor.
    NucleotideDictionary(void) {};

    /// \brief Void constructor.
    NucleotideDictionary(bool t) {};
    
    /// \brief Constructor which reads a dictionary file.
    ///
    /// The dictionary file is read immediately and only once.
    NucleotideDictionary(const char dict_name[])
      {
      read_dictionary(dict_name);
      };

   /// \brief Constructor which defines the dictionary conversions.
   ///
   /// This constructor gets rid of the need for a dictionary file.
   NucleotideDictionary(const _InternalTp& N, ///< value for unknown (char='#')
                        const _InternalTp& A, ///< value for adenine
			const _InternalTp& C, ///< value for cytosine
			const _InternalTp& G, ///< value for guanine
			const _InternalTp& T) ///< value for thymine
     {
     add('#',N);
     add('A',A); add('C',C); add('G',G); add('T',T);
     add('a',A); add('c',C); add('g',G); add('t',T);
     complementary_pair(translate('A'),translate('T'));
     complementary_pair(translate('C'),translate('G'));
     }

   /// \brief Adds new conversion rules to the dictionary.
   ///
   /// This method is simply a convenience replacement for std::map::insert.
   inline void add(const _IndexTp& smb,         ///< Symbol read from file or elsewere
                   const _InternalTp& cnv)  ///< Object to be converted to
      {
      this->insert(typename map_type::value_type(smb,cnv));
      }

   /// \brief Adds new pair of complementary symbols.
   inline void complementary_pair(const _InternalTp& fst, ///< First symbol 
                                  const _InternalTp& snd) ///< Complementary symbol
      {
      complementary_map.insert(typename complementary_map_type::value_type(fst,snd));
      complementary_map.insert(typename complementary_map_type::value_type(snd,fst));
      }

   /// \brief Conversion to complementary symbols found in complementary_map.
   ///
   /// \return Resturns the complementary symbol of type char.
   inline char complementary_symbol(const char& fst) const ///< Symbol to convert
      {
      return invert(complementary(translate(fst)));
      }

   inline _InternalTp complementary(const _InternalTp& cnv) const ///< Symbol to convert
      {
      return complementary_map.find(cnv)->second;
      }

   /// \brief Removes conversion rules from the dictionary
   inline void rm(const char& smb)   ///< Symbol to be removed
      {
      _InternalTp fst=translate(smb);
      _InternalTp snd=invert(complementary(fst));
      if (complementary_map.find(fst) != complementary_map.end()) complementary_map.erase(complementary_map.find(fst));
      if (complementary_map.find(snd) != complementary_map.end()) complementary_map.erase(complementary_map.find(snd));
      inverse_map.erase(translate(smb));
      map_type::erase(smb);
      }

    /// \brief Returns the symbolic equivalent.
    inline _IndexTp invert(const _InternalTp& cnv) const ///< object of type _InternalTp to be converted
      {
      typename DictionaryType::const_iterator dt;
      for(dt=map_type::begin(); dt != map_type::end(); dt++)
        {
	if (dt->second==cnv) return dt->first;
	}
      return _IndexTp();
      }

    /// \brief Translate a char to the _InternalTp
    ///
    /// This looks up the translation map, if not found returns '#'.
    inline _InternalTp translate(const _IndexTp& cr) const ///< char to be translated
      {
      _InternalTp result;
      typename map_type::const_iterator fnd=map_type::find(cr);
      if (fnd != map_type::end()) result=fnd->second; 
        else result=map_type::find('#')->second;
      return result;
      }

    /// \brief Extractor which loads a whole dictionary from file.
    ///
    /// Use like: file >> dict;
    friend std::istream& operator>>(std::istream &fl, DictionaryType &dict)
      {
      while (!fl.eof() && fl.good())
        {
        _IndexTp symbol; _InternalTp conversion;
        fl >> symbol >> conversion;
        dict.add(symbol,conversion);
        }
      return fl;
      }

    /// \brief Writes the contents of the dictionary to ostream.
    friend std::ostream& operator<<(std::ostream& os, const DictionaryType &dict)
      {
      typename DictionaryType::const_iterator dit;
      for(dit=dict.begin(); dit != dict.end(); dit++)
        os << dit->first << " " << dit->second << std::endl;
      return os;
      }

    /// \brief Reads a dictionary file
    void read_dictionary(const char dict_name[]) ///< name of the dictionary file
      {
      std::ifstream dictionary_file;
      dictionary_file.exceptions(std::ios_base::failbit | std::ios_base::badbit);
      try {dictionary_file.open(dict_name);}
      catch (const std::ios_base::failure& error)      {std::cerr << error.what() << __FILE__ << __LINE__ << std::endl;}
      catch (...)                       {std::cerr << "Unknown exception" << __FILE__ << __LINE__ << std::endl;}
      dictionary_file.exceptions(std::ios_base::goodbit);
      dictionary_file >> (*this);
      dictionary_file.close();
      }

    /// \brief Lists all characters known to be valid.
    char_list_type valid_characters(void)
      {
      char_list_type char_list;
      typename DictionaryType::const_iterator it;
      for (it  = map_type::begin();  it != map_type::end(); it++)
        {
        if (it->first != '#')
          {
          char_list.push_back(invert(it->second));
	  }
        }
      char_list.sort();
      char_list.unique();
      return char_list;
      }

  };
  
// //    template<>
// //      /// \brief Constructor which defines the dictionary conversions.
// //    ///
// //    /// This constructor gets rid of the need for a dictionary file.
// //    NucleotideDictionary<std::string,std::string>::NucleotideDictionary(void) ///< value for thymine
// //      {
// //      add("#","#");
// //      add("A","A"); add("C","C"); add("G","G"); add("T","T");
// //      add("a","a"); add("c","c"); add("g","g"); add("t","t");
// //      complementary_pair("A","T");
// //      complementary_pair("C","G");
// //      }

//    template<>
//     /// \brief Translate a char to the _InternalTp
//     ///
//     /// This looks up the translation map, if not found returns '#'.
//     inline std::string NucleotideDictionary<std::string,std::string>::translate(const std::string& cr) const ///< char to be translated
//       {
//       std::string result;
//       typename map_type::const_iterator fnd=map_type::find(cr);
//       if (fnd != map_type::end()) result=fnd->second; 
//         else result=map_type::find("#")->second;
//       return result;
//       }

   template<>
   /// \brief Specialization of type char.
   inline void NucleotideDictionary<char>::add(const char& smb,         ///< Symbol read from file or elsewere
                                               const char& cnv)  ///< Object to be converted to
      {
      insert(DictionaryType::value_type(smb,cnv));
      inverse_map.insert(inverse_map_type::value_type(cnv,smb));
      }

   template<>
   /// \brief Specialization of type char.
   inline char NucleotideDictionary<char>::invert(const char& cnv) const ///< object of type _InternalTp to be converted
      {
      return inverse_map.find(cnv)->second;
      }

   template<>
   /// \brief Specialization of type unsigned int.
   inline void NucleotideDictionary<unsigned int>::add(const char& smb,         ///< Symbol read from file or elsewere
                                                       const unsigned int& cnv)  ///< Object to be converted to
      {
      insert(DictionaryType::value_type(smb,cnv));
      inverse_map.insert(inverse_map_type::value_type(cnv,smb));
      }

   template<>
   /// \brief Specialization of type int.
   inline void NucleotideDictionary<int>::add(const char& smb,         ///< Symbol read from file or elsewere
                                              const int& cnv)  ///< Object to be converted to
      {
      insert(DictionaryType::value_type(smb,cnv));
      inverse_map.insert(inverse_map_type::value_type(cnv,smb));
      }

   template<>
   /// \brief Specialization of type unsigned int.
   inline char NucleotideDictionary<unsigned int>::invert(const unsigned int& cnv) const ///< object of type _InternalTp to be converted
      {
      return inverse_map.find(cnv)->second;
      }

   template<>
   /// \brief Specialization of type int.
   inline char NucleotideDictionary<int>::invert(const int& cnv) const ///< object of type _InternalTp to be converted
      {
      return inverse_map.find(cnv)->second;
      }


//    template<>
//    /// \brief Specialization of type short int.
//    inline short int NucleotideDictionary<short int>::complementary(const short int& cnv) const ///< Symbol to convert
//       {
//       return -cnv;
//       }
// 
//    template<>
//    /// \brief Specialization of type short int.
//    inline unsigned short int NucleotideDictionary<unsigned short int>::complementary(const unsigned short int& cnv) const ///< Symbol to convert
//       {
//       return 5U-cnv;
//       }

   template<>
   /// \brief Specialization of type complex<int>.
   inline std::complex<int> NucleotideDictionary<std::complex<int> >::complementary(const std::complex<int>& cnv) const ///< Symbol to convert
      {
      return conj(cnv);
      }
      

}
#endif
